structure AbstractSyntax = 
struct

type Pos = Position.PosType; 

val posToString = Position.posToString;

datatype prim_fun = eqFun | notFun | andFun | orFun | ifFun | iffFun;

datatype prim_method = claim | dn | mp | both | leftAnd | rightAnd | cd | leftEither | condRule
                     | rightEither | equiv | leftIff | rightIff | absurd | trueIntro | negRule;

datatype constant = primMethodConst of prim_method
                  | primFunConst of prim_fun
                  | propAtomConst of string
                  | trueConst 
                  | falseConst;

type ide = string;

datatype prop =  atom of string
               | trueProp
               | falseProp
               | neg of prop
               | conj of prop * prop 
               | disj of prop * prop 
               | cond of prop * prop 
               | biCond of prop * prop;

datatype exp = constExp of constant 
             | idExp of ide
             | funExp of {params: ide list, body:exp} 
             | methodExp of {params: ide list, body:ded} 
             | funAppExp of exp * phrase list
             | letExp of ((ide * phrase) list) * exp
             | fixExp of ide * exp
             | matchExp of {discriminant: exp, cases: (pattern * exp) list}
             | seqExp of exp list
     and 
         ded = methodAppDed of exp * (phrase list)
             | assumeDed of exp * ded
             | supAbDed of exp * ded
             | letDed of ((ide * phrase) list) * ded
             | byDed of exp * ded 
             | seqDed of ded list
             | matchDed of {discriminant: exp, cases: (pattern * ded) list}
    and  phrase = expression of exp | deduction of ded
    and
         pattern = idPat of ide 
                 | atomPat of string
                 | truePat 
                 | falsePat
                 | anyPat
                 | negPat of pattern
                 | conjPat of pattern * pattern
                 | disjPat of pattern * pattern
                 | condPat of pattern * pattern
                 | biCondPat of pattern * pattern;

fun patIds(idPat(id)) = [id]
  | patIds(negPat(pat)) = patIds(pat)
  | patIds(conjPat(pat1,pat2)) = patIds(pat1)@patIds(pat2)
  | patIds(disjPat(pat1,pat2)) = patIds(pat1)@patIds(pat2)
  | patIds(condPat(pat1,pat2)) = patIds(pat1)@patIds(pat2)
  | patIds(biCondPat(pat1,pat2)) = patIds(pat1)@patIds(pat2)
  | patIds(_) = [];

fun patSize(negPat(p)) = 1 + patSize(p)
  | patSize(conjPat(p1,p2)) = 1 + patSize(p1) + patSize(p2)
  | patSize(disjPat(p1,p2)) = 1 + patSize(p1) + patSize(p2)
  | patSize(condPat(p1,p2)) = 1 + patSize(p1) + patSize(p2)
  | patSize(biCondPat(p1,p2)) = 1 + patSize(p1) + patSize(p2)
  | patSize(_) = 1;

fun propExp(atom(str)) = constExp(propAtomConst(str))
  | propExp(trueProp) = constExp(trueConst)
  | propExp(falseProp) = constExp(falseConst)
  | propExp(neg(P)) = funAppExp(constExp(primFunConst(notFun)),[expression(propExp(P))])
  | propExp(conj(P1,P2)) = funAppExp(constExp(primFunConst(andFun)),
                                     [expression(propExp(P1)),expression(propExp(P2))])
  | propExp(disj(P1,P2)) = funAppExp(constExp(primFunConst(orFun)),
                                     [expression(propExp(P1)),expression(propExp(P2))])
  | propExp(cond(P1,P2)) = funAppExp(constExp(primFunConst(ifFun)),
                                     [expression(propExp(P1)),expression(propExp(P2))])
  | propExp(biCond(P1,P2)) = funAppExp(constExp(primFunConst(iffFun)),
                                     [expression(propExp(P1)),expression(propExp(P2))]);

fun expToProp(constExp(propAtomConst(str))) = atom(str)
  | expToProp(constExp(trueConst)) = trueProp
  | expToProp(constExp(falseConst)) = falseProp
  | expToProp(funAppExp(constExp(primFunConst(notFun)),[expression(E)])) = neg(expToProp(E))
  | expToProp(funAppExp(constExp(primFunConst(andFun)),[expression(E1),expression(E2)])) = 
      conj(expToProp(E1),expToProp(E2))
  | expToProp(funAppExp(constExp(primFunConst(orFun)),[expression(E1),expression(E2)])) = 
      disj(expToProp(E1),expToProp(E2))
  | expToProp(funAppExp(constExp(primFunConst(ifFun)),[expression(E1),expression(E2)])) = 
      cond(expToProp(E1),expToProp(E2))
  | expToProp(funAppExp(constExp(primFunConst(iffFun)),[expression(E1),expression(E2)])) = 
      biCond(expToProp(E1),expToProp(E2));

fun phraseToProp(expression(e)) = expToProp(e);

fun getProp(phr) = phraseToProp(phr);

fun getProps([]) = []
  | getProps(phr::more) = getProp(phr)::getProps(more);


datatype user_input = definition of string * phrase 
                    | phraseInput of phrase 
                    | assertion of exp 
                    | minimalPrintOuts 
                    | noTypes 
                    | load of string;

exception LexError of string * (Pos option);

fun printPrimFun(eqFun) = "="
  | printPrimFun(notFun) = "~"
  | printPrimFun(andFun) = "&"
  | printPrimFun(orFun) = "\\/"
  | printPrimFun(ifFun) = "==>"
  | printPrimFun(iffFun) = "<==>";

fun printPrimMethod(claim) = "claim"
  | printPrimMethod(dn) = "dn"
  | printPrimMethod(mp) = "mp"
  | printPrimMethod(both) = "both"
  | printPrimMethod(leftAnd) = "left-and"
  | printPrimMethod(rightAnd) = "right-and"
  | printPrimMethod(cd) = "cd"
  | printPrimMethod(leftEither) = "left-either"
  | printPrimMethod(rightEither) = "right-either"
  | printPrimMethod(equiv) = "equiv"
  | printPrimMethod(leftIff) = "left-iff"
  | printPrimMethod(rightIff) = "right-iff"
  | printPrimMethod(absurd) = "absurd"
  | printPrimMethod(condRule) = "cond"
  | printPrimMethod(negRule) = "neg"
  | printPrimMethod(trueIntro) = "true-intro";


fun printProp(atom(str)) = str
  | printProp(trueProp) = "true"
  | printProp(falseProp) = "false"
  | printProp(neg(P)) = "(~ "^printProp(P)^")"
  | printProp(conj(P1,P2)) = "(& "^printProp(P1)^" "^printProp(P2)^")"
  | printProp(disj(P1,P2)) = "(\\/ "^printProp(P1)^" "^printProp(P2)^")"
  | printProp(cond(P1,P2)) = "(==> "^printProp(P1)^" "^printProp(P2)^")"
  | printProp(biCond(P1,P2)) = "(<==> "^printProp(P1)^" "^printProp(P2)^")"


fun primMethToString(claim) = "claim"
  | primMethToString(dn) = "double-negation"
  | primMethToString(mp) = "modus-ponens"
  | primMethToString(both) = "both"
  | primMethToString(leftAnd) = "left-and"
  | primMethToString(rightAnd) = "right-and"
  | primMethToString(cd) = "constructive-dilemma"
  | primMethToString(leftEither) = "left-either"
  | primMethToString(rightEither) = "right-either"
  | primMethToString(equiv) = "equiv"
  | primMethToString(leftIff) = "left-iff"
  | primMethToString(rightIff) = "right-iff"
  | primMethToString(absurd) = "absurd"
  | primMethToString(negRule) = "neg"
  | primMethToString(condRule) = "cond"
  | primMethToString(trueIntro) = "true-intro";

fun lift(seqDed([D])) = D
  | lift(seqDed(D::rest)) = seqDed(lift(D)::(map lift rest))
  | lift(assumeDed(E,D)) = assumeDed(E,lift D)
  | lift(supAbDed(E,D)) = supAbDed(E,lift D)
  | lift(D) = D;

fun unparseAlphaDed(methodAppDed(constExp(primMethodConst(m)),args)) = 
      primMethToString(m)^" "^Basic.printListStr(map phraseToProp args,printProp)
  | unparseAlphaDed(assumeDed(e,d)) = "assume "^printProp(expToProp(e))^" in\n  "^unparseAlphaDed(d)
  | unparseAlphaDed(supAbDed(e,d)) = "suppose-absurd "^printProp(expToProp(e))^" in\n  "^unparseAlphaDed(d)
  | unparseAlphaDed(seqDed(deds)) = "["^Basic.printListStrWithSep(deds,unparseAlphaDed,";\n")^"]"
  | unparseAlphaDed(_) = "";


end; (* of structure AbSyntax *)
