structure Types = 
struct

structure A = AbstractSyntax;

datatype cnd_type = typeVar of int | propType | funType of cnd_type list * cnd_type 
                  | methType of cnd_type list | ugenType of int * cnd_type;

type sub = int -> cnd_type;

val empty_sub = fn i => typeVar(i);

fun extendSub(sub,i,t) = (fn j => if j = i then t else sub(j));

fun singleton(i,t) = extendSub(empty_sub,i,t);

fun freeTVars(typeVar(i)) = [i]
  | freeTVars(propType) = [] 
  | freeTVars(funType(arg_types,res_type)) = List.foldl op@ [] (map freeTVars (res_type::arg_types))
  | freeTVars(methType(arg_types)) = List.foldl op@ [] (map freeTVars arg_types)
  | freeTVars(ugenType(i,body)) = Basic.remove(i,freeTVars(body));

fun occursIn(i,t) = Basic.isMember(i,freeTVars(t));

fun generalize(t,[]) = t 
  | generalize(t,v::rest) = ugenType(v,generalize(t,rest));

fun lift sub = (fn typeVar(i) => sub(i)
                 | propType => propType
                 | funType(arg_types,res_type) => funType(map (lift sub) arg_types,(lift sub) res_type)
                 | methType(arg_types) => methType(map (lift sub) arg_types)
                 | ugenType(i,body) => ugenType(i,(lift (Semant.extend(sub,[(i,typeVar(i))])) body)));

fun printType(typeVar(i)) = "T"^Int.toString(i) 
  | printType(propType) = "prop" 
  | printType(funType(arg_types,res_type)) = 
             if List.length(arg_types) > 1 then
                "(("^Basic.printListStrWithSep(arg_types,printType," * ")^") --> "^
                                             printType(res_type)^")"
             else
                "("^Basic.printListStrWithSep(arg_types,printType," * ")^" --> "^printType(res_type)^")"
  | printType(methType(arg_types)) = 
          if List.length(arg_types) > 1 then
             "(("^Basic.printListStrWithSep(arg_types,printType," * ")^") -->> prop)"
          else
             "("^Basic.printListStrWithSep(arg_types,printType," * ")^" -->> prop)"
  | printType(ugenType(i,body)) = "(FORALL T"^Int.toString(i)^" "^printType(body)^")";

fun printTypeList(types) = Basic.printListStr(types,printType);

fun compose(sub2,sub1:sub) = fn i => lift sub2 (sub1 i);

exception UnifyError;

exception UnBoundIdentifier;

fun uniFail() = raise UnifyError;

fun unify(propType,propType) = empty_sub
  | unify(typeVar(i),t) = if not(occursIn(i,t)) then extendSub(empty_sub,i,t) else
                          (case t of
                              typeVar(_) => empty_sub
                            | _ => uniFail())
  | unify(s,t as typeVar(i)) = unify(t,s)
  | unify(funType(arg_types1,res_type1),funType(arg_types2,res_type2)) = 
           unifyLst(arg_types1@[res_type1],arg_types2@[res_type2])
  | unify(methType(arg_types1),methType(arg_types2)) = unifyLst(arg_types1,arg_types2)
  | unify(_) = uniFail()
and
    unifyLst([],[]) = empty_sub
  | unifyLst(t1::rest1,t2::rest2) = let val sub1 = unify(t1,t2)  
                                        val sub1' = lift sub1
                                        val sub2 = unifyLst(map sub1' rest1,map sub1' rest2)
                                    in 
                                       compose(sub2,sub1)
                                    end
  | unifyLst(_) = uniFail();

type type_env = (A.ide * cnd_type) list;

fun printTypeEnv(te) = print(Basic.printListStr(te, fn (id,t) => (id^": "^(printType t))));

fun lookUp(id,[]) = NONE
  | lookUp(id,(id',t)::more) = if id = id' then SOME(t) else lookUp(id,more);

val empty_type_env: type_env = [];

fun appSubToTypeEnv(_,[]) = []
  | appSubToTypeEnv(sub,(id,t)::rest) = let val res = (lift sub) t                                   
                                        in
                                          (id,res)::(appSubToTypeEnv(sub,rest))
                                        end;

fun extendTypeEnv(tenv,id,t) = (id,t)::tenv;

fun augmentTypeEnv(te,te') = te@te';

fun freeTEnvVars(te) = List.foldl op@ [] (map (fn (id,t) => freeTVars(t)) te);

fun Delta(pat) = map (fn id => (id,propType)) (A.patIds pat);

fun tc(A.primMethodConst(A.claim)) = methType([propType])
  | tc(A.primMethodConst(A.dn)) = methType([propType])
  | tc(A.primMethodConst(A.negRule)) = methType([propType])
  | tc(A.primMethodConst(A.condRule)) = methType([propType,propType])
  | tc(A.primMethodConst(A.leftAnd)) = methType([propType])
  | tc(A.primMethodConst(A.rightAnd)) = methType([propType])
  | tc(A.primMethodConst(A.leftIff)) = methType([propType])
  | tc(A.primMethodConst(A.rightIff)) = methType([propType])
  | tc(A.primMethodConst(A.mp)) = methType([propType,propType])
  | tc(A.primMethodConst(A.both)) = methType([propType,propType])
  | tc(A.primMethodConst(A.leftEither)) = methType([propType,propType])
  | tc(A.primMethodConst(A.rightEither)) = methType([propType,propType])
  | tc(A.primMethodConst(A.equiv)) = methType([propType,propType])
  | tc(A.primMethodConst(A.absurd)) = methType([propType,propType])
  | tc(A.primMethodConst(A.cd)) = methType([propType,propType,propType])
  | tc(A.primMethodConst(A.trueIntro)) = methType([])
  | tc(A.primFunConst(A.eqFun)) = funType([propType,propType],propType)
  | tc(A.primFunConst(A.andFun)) = funType([propType,propType],propType)
  | tc(A.primFunConst(A.orFun)) = funType([propType,propType],propType)
  | tc(A.primFunConst(A.ifFun)) = funType([propType,propType],propType)
  | tc(A.primFunConst(A.iffFun)) = funType([propType,propType],propType)
  | tc(A.primFunConst(A.notFun)) = funType([propType],propType)
  | tc(_) = propType;

fun typeConstant(c,_) = (tc(c),empty_sub);

fun inferType(phr,tenv) = 
 let val counter = ref(1)
     fun freshType() = let val i = !counter in (counter := !counter + 1; typeVar(i)) end
     fun freshInstance(ugenType(i,t)) = (lift (singleton(i,freshType()))) (freshInstance t)
       | freshInstance(t) = t
     fun VDed(A.methodAppDed(e,args),gamma) = 
            let val (t,theta) = VExp(e,gamma) 
                val (arg_types,theta1) = VPhraseLst(args,map (fn _ => gamma) args,theta)
                val theta2 = unify((lift theta1) t,methType(arg_types))
            in
                (propType,compose(theta2,theta1))
            end
       | VDed(A.assumeDed(hyp,body),te) = 
           let val (t,theta1) = VExp(hyp,te)
               val theta2 = unify(t,propType)
               val (tau',theta3) = VDed(body,appSubToTypeEnv(compose(theta2,theta1),te))
               val theta4 = unify(tau',propType)
           in
             (propType,compose(theta4,compose(theta3,compose(theta2,theta1))))
           end
       | VDed(A.supAbDed(hyp,body),te) = 
           let val (t,theta1) = VExp(hyp,te)
               val theta2 = unify(t,propType)
               val (tau',theta3) = VDed(body,appSubToTypeEnv(compose(theta2,theta1),te))
               val theta4 = unify(tau',propType)
           in
             (propType,compose(theta4,compose(theta3,compose(theta2,theta1))))
           end
       | VDed(A.letDed([],body),te) = VDed(body,te)
       | VDed(A.letDed((id,phr)::rest,body),te) = 
            let val (t1,theta1) = VPhrase(phr,te)
                val te' = appSubToTypeEnv(theta1,te)
                val fvars = Basic.removeDuplicatesEq(
                             Basic.removeSeveral(freeTEnvVars(te'),freeTVars(t1)),op=)
                val ut = generalize(t1,fvars)
                val te'' = extendTypeEnv(te',id,ut)
                val (td,theta2) = VDed(A.letDed(rest,body),te'')
                val theta3 = unify(td,propType)
            in
              (propType,compose(theta3,compose(theta2,theta1)))
            end
       | VDed(A.byDed(e,d),te) = 
            let val (t1,theta1) = VDed(d,te)
                val theta2 = unify(t1,propType)
                val (t2,theta3) = VExp(e,appSubToTypeEnv(compose(theta2,theta1),te))
                val theta4 = unify(t2,propType)
            in
               (propType,compose(theta4,compose(theta3,compose(theta2,theta1))))
            end
       | VDed(A.matchDed({discriminant, cases}),te) = 
            let val (t,theta1) = VExp(discriminant,te)
                val theta2 = unify(t,propType)
                val (pats,deds) = ListPair.unzip(cases)
                val gamma_lst = map Delta pats                  
                val gamma_lst' = map (fn te' => augmentTypeEnv(te',te)) gamma_lst
                val (case_types,theta) = VPhraseLst(map A.deduction deds,gamma_lst',compose(theta2,theta1))
                val theta' = unifyLst(map (lift theta) case_types,map (fn _ => propType) cases)
            in     
               (propType,compose(theta',theta))
            end
       | VDed(A.seqDed(deds),te) = 
            let val (types,theta) = VPhraseLst(map A.deduction deds,map (fn _ => te) deds,empty_sub)
                val theta' = unifyLst(types,map (fn _ => propType) types)
            in
               (propType,compose(theta',theta))
            end
     and VExp(A.constExp(c),te) = typeConstant(c,te)
       | VExp(A.idExp(id),te) = (case lookUp(id,te) of 
                                     SOME(t) => (freshInstance(t),empty_sub)
                                   | _ => raise UnBoundIdentifier)
       | VExp(A.funExp({params,body}),te) = 
           let val new_types = map (fn _ => freshType()) params
               val (t,theta) = VExp(body,augmentTypeEnv(ListPair.zip(params,new_types),te))
           in 
              ((lift theta) (funType(new_types,t)),theta)
           end
       | VExp(A.methodExp({params,body}),te) = 
           let val new_types = map (fn _ => freshType()) params
               val (t,theta) = VDed(body,augmentTypeEnv(ListPair.zip(params,new_types),te))
               val theta1 = unify(t,propType)
               val theta2 = compose(theta1,theta)
           in 
              ((lift theta2) (methType(new_types)),theta2)
           end
       | VExp(A.funAppExp(e,phrases),te) = 
           let val (t,theta) = VExp(e,te)
               val (arg_types,theta1) = VPhraseLst(phrases,map (fn _ => te) phrases,theta)
               val ft = freshType()
               val theta2 = unify((lift theta1) t, funType(arg_types,ft))
           in
              ((lift theta2) ft,compose(theta2,theta1))
           end
       | VExp(A.letExp([],body),te) = VExp(body,te)
       | VExp(A.letExp((id,phr)::rest,body),te) = 
            let val (t1,theta1) = VPhrase(phr,te)
                val te' = appSubToTypeEnv(theta1,te)
                val fvars = Basic.removeDuplicatesEq(
                               Basic.removeSeveral(freeTEnvVars(te'),freeTVars(t1)),op=)
                val ut = generalize(t1,fvars)
                val te'' = extendTypeEnv(te',id,ut)
                val (te,theta2) = VExp(A.letExp(rest,body),te'')
            in
              (te,compose(theta2,theta1))
            end
       | VExp(A.matchExp({discriminant, cases}),te) = 
            let val (t,theta1) = VExp(discriminant,te)
                val theta2 = unify(t,propType)
                val (pats,exps) = ListPair.unzip(cases)
                val gamma_lst = map Delta pats                  
                val gamma_lst' = map (fn te' => augmentTypeEnv(te',te)) gamma_lst
                val (case_types,theta) = VPhraseLst(map A.expression exps,gamma_lst',compose(theta2,theta1))
                val theta' = unifyLst(map (lift theta) (List.take(case_types,List.length(case_types) - 1)),
                                      map (lift theta) (tl case_types))
            in     
               ((lift theta') (hd case_types),compose(theta',theta))
            end
       | VExp(A.fixExp(id,body),te) = 
            let val ft = freshType()
                val (t,theta) = VExp(body,extendTypeEnv(te,id,ft))
                val theta1 = unify(t,(lift theta) ft)
                val theta2 = compose(theta1,theta)
            in
               ((lift theta2) ft,theta2)
            end
       | VExp(A.seqExp(exps),te) = 
            let val (types,theta) = VPhraseLst(map A.expression exps,map (fn _ => te) exps,empty_sub)
            in
               ((lift theta) (hd (rev types)),theta)
            end
     and VPhrase(A.expression(e),te) = VExp(e,te)
       | VPhrase(A.deduction(d),te) = VDed(d,te)
     and VPhraseLst(phrases,tenvs,sub) = 
          let fun h([],[],theta,types) = (rev(types),theta)
                | h(phr::more_phrases,te::more_tes,theta,types) = 
                       let val (t,theta') = VPhrase(phr,appSubToTypeEnv(theta,te))
                       in
                         h(more_phrases,more_tes,compose(theta',theta),t::types)
                       end
          in
             h(phrases,tenvs,sub,[])
          end
     val (t,_) = VPhrase(phr,tenv)
  in
    t
  end;
                                               
end; (* of structure Types *)
